/****************************************************************************/
/*                                                                          */
/*  Dc_Graphic.c :  Module de gestion des ressources graphiques.            */
/*                                                                          */
/****************************************************************************/
/*  Auteur : Franois GAUTHIER   *   Cooperteam   *   Date : Septembre 2010 */
/****************************************************************************/

#include <windows.h>

#include <stdio.h>
#include <stdlib.h>

#include "Dc_Memory.h"
#include "Dc_Shared.h"
#include "Dc_Graphic.h"
#include "Dc_Window.h"
#include "Dc_Gif.h"
#include "resource.h"

struct thread_parameter          /* Paramtre ddi au Thread */
{
  DWORD process_id;             /* ID identifiant ce Process */
  DWORD thread_id;              /* ID identifiant ce Thread */

  struct window_settings *current_settings;

  char file_path[2048];
  int file_format;

  int picture_width;
  int picture_height;
  
  unsigned char *bitmap_data;
  
  int nb_color;
  int palette_ind[257];
};

DWORD CreateOnePictureOnDiskThread(void *);
void mem_free_tparam(struct thread_parameter *);
int CreateGraphicBmp(char *,int,int,unsigned char *,int[]);
int LoadGraphicBmp(char *,int *,int *,unsigned char **);
int ConvertTrueColorTo256(int,int,unsigned char[],unsigned char **,int[],int[],int[]);

/*****************************************************************/
/*  CreateOnePictureOnDisk() :  Cration d'une image sur disque. */
/*****************************************************************/
void CreateOnePictureOnDisk(struct window_settings *current_settings, char *file_path, int file_format, int picture_width, int picture_height, unsigned char *bitmap_data, int nb_color, int *palette_ind)
{
  HANDLE thread_handle;
  DWORD thread_id;
  struct thread_parameter *tparam;

  /* Allocation de la structure thread_parameter */
  tparam = (struct thread_parameter *) calloc(1,sizeof(struct thread_parameter));
  if(tparam == NULL)
    return;

  /** Remplit la structure **/
  tparam->current_settings = current_settings;
  strcpy(tparam->file_path,file_path);
  tparam->file_format = file_format;
  tparam->picture_width = picture_width;
  tparam->picture_height = picture_height;
  tparam->bitmap_data = bitmap_data;
  tparam->nb_color = nb_color;
  memcpy(&tparam->palette_ind[0],&palette_ind[0],257*sizeof(int));

  /** Lancement du Thread en charge de la cration de l'image sur disque **/
  thread_handle = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)CreateOnePictureOnDiskThread,tparam,0,&thread_id);
  if(thread_handle == NULL)
    {
      /* Libration mmoire */
      mem_free_tparam(tparam);
    }
}


/*****************************************************************************************/
/*  CreateOnePictureOnDiskThread() :  Traite la cration de l'image via 1 Thread ddi.  */
/*****************************************************************************************/
DWORD CreateOnePictureOnDiskThread(void *thread_param)
{
  int error;
  unsigned char *data_256;
  int pal_red[256];
  int pal_green[256];
  int pal_blue[256];
  char buffer_error[1024];
  struct thread_parameter *tparam = (struct thread_parameter *)thread_param;

  /* ID identifiant ce Thread */
  tparam->process_id = GetCurrentProcessId();
  tparam->thread_id = GetCurrentThreadId();

  /*** On oriente vers le format de fichier souhait ***/
  if(tparam->file_format == FILE_FORMAT_BMP)
    {
      /** Cration de l'image BMP sur disque **/
      error = CreateGraphicBmp(tparam->file_path,tparam->picture_height,tparam->picture_width,tparam->bitmap_data,tparam->palette_ind);
      if(error == ERR_GRAPHIC_FOPEN)
        strcpy(buffer_error,"CreatGraphicBmp fopen() failed!");
      else if( error == ERR_GRAPHIC_WRITEFILE)
        strcpy(buffer_error,"CreatGraphicBmp fwrite() failed!");
      else if( error == ERR_GRAPHIC_BADFORMAT)
        strcpy(buffer_error,"File size is NULL!");

      /* Libration mmoire */
      mem_free_tparam(tparam);
    }
  else    /* GIF */
    {
      /** Conversion en 8 bits de l'image **/
      error = ConvertTrueColorTo256(tparam->picture_height,tparam->picture_width,tparam->bitmap_data,&data_256,&pal_red[0],&pal_green[0],&pal_blue[0]);
      if(error == TOO_MANY_COLOR)
        {
          free(tparam->bitmap_data);
          strcpy(buffer_error,"Too many colors for GIF format!");
          return(0);
        }
      else if(error == ERR_ALLOC)
        {
          free(tparam->bitmap_data);
          strcpy(buffer_error,"Memory allocation failed!");
          return(0);
        }

      /** Cration du fichier GIF sur disque **/
      GIFWriteFileFrom256Color(data_256,tparam->file_path,tparam->picture_width,tparam->picture_height,0,&pal_red[0],&pal_green[0],&pal_blue[0]);

      /* Libration mmoire */
      free(data_256);
      mem_free_tparam(tparam);
    }

  /* OK */
  return(1);
}


/******************************************************************************/
/*  mem_free_tparam() : Libration mmoire de la structure thread_parameter.  */
/******************************************************************************/
void mem_free_tparam(struct thread_parameter *tparam)
{
  if(tparam)
    {
      free(tparam->bitmap_data);
      free(tparam);
    }
}


/*********************************************************/
/*  LoadGraphicBmp() : Charge un fichier au format BMP.  */
/*********************************************************/
static int LoadGraphicBmp(char *file_path, int *height_rtn, int *width_rtn, unsigned char **picture_rtn)
{
  int i, j;
  int error;
  int width, height, color_offset, color, color_1, color_2;
  int offset_src, offset_dst, nb_point;
  long file_size;
  FILE *fd;
  unsigned char *buffer;
  unsigned char *picture;
  unsigned char *clipboard;
  BITMAPFILEHEADER bmfh;
  BITMAPINFOHEADER bmih;
  memset(&bmfh,0,sizeof(BITMAPFILEHEADER));
  memset(&bmih,0,sizeof(BITMAPINFOHEADER));

  /** Lecture du fichier en mmoire **/
  /* Ouverture du fichier */
  fd = fopen(file_path,"r");
  if(fd == NULL)
    {
      fclose(fd);
      return(ERR_GRAPHIC_FOPEN);
    }  
  /* Dtermine la taille du fichier */
  fseek(fd,0L,SEEK_END);
  file_size = ftell(fd);
  fseek(fd,0L,SEEK_SET);

  /* Cration d'un buffer de la taille du fichier */
  buffer = (unsigned char *) calloc(1,file_size+1);
  if(buffer == NULL)
    {
      fclose(fd);
      return(ERR_GRAPHIC_BADFORMAT);
    }

  /* Rcupration de l'entte */
  memcpy(&bmfh,buffer,sizeof(BITMAPFILEHEADER));
  memcpy(&bmih,&buffer[sizeof(BITMAPFILEHEADER)],sizeof(BITMAPINFOHEADER));
  width = bmih.biWidth;
  height = bmih.biHeight;
  color_offset = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
  offset_src = bmfh.bfOffBits;
  offset_dst = 0;
  nb_point = 0;

  /* Vrification du format */
  if(bmfh.bfType != 19778 || (long)bmfh.bfSize != file_size)
    return(ERR_GRAPHIC_BADFORMAT);

  /* Allocation de la picture */
  picture =(unsigned char *) calloc(1,width*height*3);
  if(picture == NULL)
    return(error);

  /** On dcompresse **/
  if(bmih.biCompression == 1)   /* BI_RLE8 : 256 couleurs */
    {
      for(i=0;;)
        {
          if(buffer[offset_src+i] == 0x00)
            {
              if(buffer[offset_src+i+1] == 0x00)  /* Fin de ligne */
                  i+=2;
              else if(buffer[offset_src+i+1] == 0x01)  /* Fin de bitmap */
                break;
              else if(buffer[offset_src+i+1] == 0x02)  /* Offset */
                {
                  offset_dst += ((int)buffer[offset_src+i+2] + (int)buffer[offset_src+i+3]*width )*3;
                  i+=4;
                }
              else  /* On recopie les octets */
                {
                  for(j=0; j<(int)buffer[offset_src+i+1]; j++)
                    {
                      color = buffer[offset_src+i+2+j];
                      picture[offset_dst+0] = buffer[color_offset + 4*color + 2];
                      picture[offset_dst+1] = buffer[color_offset + 4*color + 1];
                      picture[offset_dst+2] = buffer[color_offset + 4*color + 0];
                      offset_dst+=3;
                    }
                  i += (2+(int)buffer[offset_src+i+1]+(int)buffer[offset_src+i+1]%2);
                }
            }
          else
            {
              /* Rptition de points */
              for(j=0; j< (int)buffer[offset_src+i]; j++)
                {
                  color = buffer[offset_src+i+1];
                  picture[offset_dst+0] = buffer[color_offset + 4*color + 2];
                  picture[offset_dst+1] = buffer[color_offset + 4*color + 1];
                  picture[offset_dst+2] = buffer[color_offset + 4*color + 0];
                  offset_dst+=3;
                }
              i += 2;
            }
        }
    }
  if(bmih.biCompression == 2)   /* BI_RLE4 : 4 couleurs */
    {
      for(i=0;;)
        {
          if(buffer[offset_src+i] == 0x00)
            {
              if(buffer[offset_src+i+1] == 0x00)  /* Fin de ligne */
                {
                  i+=2;
                  nb_point = 0;
                }
              else if(buffer[offset_src+i+1] == 0x01)  /* Fin de bitmap */
                break;
              else if(buffer[offset_src+i+1] == 0x02)  /* Offset */
                {
                  offset_dst += ((int)buffer[offset_src+i+2] + (int)buffer[offset_src+i+3]*width)*3;
                  nb_point = (int)buffer[offset_src+i+2];
                  i+=4;
                }
              else  /* On recopie les octets */
                {
                  for(j=0; j<(int)buffer[offset_src+i+1]; j++)
                    {
                      if(!(j%2))
                        color = (buffer[offset_src+i+2+j/2] & 0xF0) >> 4;
                      else
                        color = (buffer[offset_src+i+2+j/2] & 0x0F);

                      if(nb_point < width)
                        {
                          picture[offset_dst+0] = buffer[color_offset + 4*color + 2];
                          picture[offset_dst+1] = buffer[color_offset + 4*color + 1];
                          picture[offset_dst+2] = buffer[color_offset + 4*color + 0];
                          offset_dst+=3;
                          nb_point++;
                        }
                    }
                  i += (2+((int)buffer[offset_src+i+1]+1)/2+(int)((buffer[offset_src+i+1]+1)/2)%2);
                }
            }
          else
            {
              /* Rptition de points */
              color_1 = (buffer[offset_src+i+1] & 0xF0) >> 4;
              color_2 = (buffer[offset_src+i+1] & 0x0F);
              for(j=0; j< (int)buffer[offset_src+i]; j++)
                {
                  if(nb_point < width)
                    {
                      if(!(j%2))
                        {
                          picture[offset_dst+0] = buffer[color_offset + 4*color_1 + 2];
                          picture[offset_dst+1] = buffer[color_offset + 4*color_1 + 1];
                          picture[offset_dst+2] = buffer[color_offset + 4*color_1 + 0];
                        }
                      else
                        {
                          picture[offset_dst+0] = buffer[color_offset + 4*color_2 + 2];
                          picture[offset_dst+1] = buffer[color_offset + 4*color_2 + 1];
                          picture[offset_dst+2] = buffer[color_offset + 4*color_2 + 0];
                        }
                      offset_dst+=3;
                      nb_point++;
                    }
                }
              i += 2;
            }
        }
    }
  else if(bmih.biCompression == 0)  /* BI_RGB */
    {
      /** Image non compresse **/
      if(bmih.biBitCount == 1)
        {
          /** On va chercher les couleurs dans la palette **/
          for(i=0; i<height; i++)
            {
              for(j=0; j<width; j++)
                {
                  color = ((buffer[offset_src+j/8] >> 7)-(j%8)) & 0x01;
                  picture[offset_dst+3*j+0] = buffer[color_offset + 4*color + 2];
                  picture[offset_dst+3*j+1] = buffer[color_offset + 4*color + 1];
                  picture[offset_dst+3*j+2] = buffer[color_offset + 4*color + 0];
                }
              offset_src += (width + ((width%32) == 0 ? 0 :32-(width%32)))/8;
              offset_dst += width*3;
            }
        }
      else if(bmih.biBitCount == 4)
        {
          /** On va chercher les couleurs dans la palette **/
          for(i=0; i<height; i++)
            {
              for(j=0; j<width; j++)
                {
                  color = (j%2) ? (0x0F & buffer[offset_src+j/2]) : (0xF0 & buffer[offset_src+j/2]) >> 4;
                  picture[offset_dst+3*j+0] = buffer[color_offset + 4*color + 2];
                  picture[offset_dst+3*j+1] = buffer[color_offset + 4*color + 1];
                  picture[offset_dst+3*j+2] = buffer[color_offset + 4*color + 0];
                }
              offset_src += (width + ((width%8 == 0) ? 0 : 8-(width%8)))/2;
              offset_dst += width*3;
            }
        }
      else if(bmih.biBitCount == 8)
        {
          /** On va chercher les couleurs dans la palette **/
          for(i=0; i<height; i++)
            {
              for(j=0; j<width; j++)
                {
                  color = buffer[offset_src+j];
                  picture[offset_dst+3*j+0] = buffer[color_offset + 4*color + 2];
                  picture[offset_dst+3*j+1] = buffer[color_offset + 4*color + 1];
                  picture[offset_dst+3*j+2] = buffer[color_offset + 4*color + 0];
                }
              offset_src += width + ((width%4 == 0) ? 0 : 4-(width%4));
              offset_dst += width*3;
            }
        }
      else if(bmih.biBitCount == 24)
        {
          /** On rcupre directement les points **/
          for(i=0; i<height; i++)
            {
              for(j=0; j<width; j++)
                {
                  picture[offset_dst+3*j+0] = buffer[offset_src+3*j+2];
                  picture[offset_dst+3*j+1] = buffer[offset_src+3*j+1];
                  picture[offset_dst+3*j+2] = buffer[offset_src+3*j+0];
                }
              offset_src += (width*3 + ((width*3)%4 == 0 ? 0 : 4-(width*3)%4));
              offset_dst += width*3;
            }
        }
    }

  /** On retourne l'image **/
  for(i=0; i<height; i++)
    memcpy(&clipboard[(height-i-1)*width*3],&picture[i*width*3],width*3);
  memcpy(picture,clipboard,width*height*3);

  /* Libration mmoire */
  free(buffer);
  free(clipboard);
  return(0);
}


/****************************************************************************/
/*  ConvertTrueColorTo256() : Conversion du true color en 256 couleurs GIF. */
/****************************************************************************/
int ConvertTrueColorTo256(int window_height, int window_width, unsigned char *data_tc,  unsigned char **data_256_rtn, int pal_red[256], int pal_green[256], int pal_blue[256])
{
  int i,j;
  int offset, nb_real_color;
  int first, last, x, color_ind;
  int palette_ind[256];
  unsigned char red, green, blue;
  unsigned char *data_256;

  /* Init Palette */
  memset(&pal_red[0],0,256*sizeof(int));
  memset(&pal_green[0],0,256*sizeof(int));
  memset(&pal_blue[0],0,256*sizeof(int));

  /** Cration de la palette **/
  nb_real_color = 0;
  for(j=0; j<window_height; j++)
    for(i=0; i<window_width; i++)
      {
        /* Couleur du point */
        offset = (i+j*window_width)*3;
        red = data_tc[offset];
        green = data_tc[offset+1];
        blue = data_tc[offset+2];
        color_ind = 65536*red + 256*green + blue;

        /** On insert/recherche la couleur (dichotomie) **/
        if(i == 0 && j == 0)
          {
            palette_ind[0] = color_ind;
            nb_real_color = 1;
          }
        else
          {
            for(first=0,last=nb_real_color; last>=first;)
              {
                x = (first+last)/2;
                if(color_ind < palette_ind[x])
                  last = x - 1;
                else
                  first = x + 1;

                if(color_ind == palette_ind[x])
                  break;
              }

            /* A t'on trouv la couleur ou doit on l'ajouter */
            if(color_ind != palette_ind[x])
              {
                /* Gestion des images > 256 couleurs */
                if(nb_real_color == 256)
                  return(TOO_MANY_COLOR);

                /* Dcalage du tableau */
                if(x < nb_real_color)
                  if(palette_ind[x] < color_ind)
                    x++;
                memmove(&palette_ind[x+1],&palette_ind[x],(nb_real_color-x)*sizeof(int));
                palette_ind[x] = color_ind;
                nb_real_color++;
              }
          }
      }
  /* Palette spare RGB */
  for(i=0; i<nb_real_color; i++)
    {
      pal_red[i] = (palette_ind[i]&0x00FF0000)>>16;
      pal_green[i] = (palette_ind[i]&0x0000FF00)>>8;
      pal_blue[i] = (palette_ind[i]&0x000000FF);
    }

  /* Allocation mmoire de l'image en 256 couleurs */
  data_256 = (unsigned char *) calloc(1,window_height*window_width);
  if(data_256 == NULL)
    return(ERR_ALLOC);

  /** Points **/
  for(j=0; j<window_height; j++)
    for(i=0; i<window_width; i++)
      {
        offset = (i+j*window_width)*3;
        red = data_tc[offset];
        green = data_tc[offset+1];
        blue = data_tc[offset+2];
        color_ind = 65536*red + 256*green + blue;

        /** Recherche dichotomique de la valeur **/
        for(first=0,last=nb_real_color; last>=first;)
          {
            x = (first+last)/2;
            if(color_ind < palette_ind[x])
              last = x - 1;
            else
              first = x + 1;

            if(color_ind == palette_ind[x])
              break;
          }

        /* Cration du point */
        data_256[i+j*window_width] = (unsigned char) x;
      }
 
  /** Renvoi les points et la palette **/  
  *data_256_rtn = data_256; 
  return(0);
}


/************************************************************/
/*  CreateGraphicBmp() : Cration du fichier au format Bmp. */
/************************************************************/
int CreateGraphicBmp(char *file_path, int height, int width, unsigned char *picture, int *palette_indice)
{
  FILE *fd;
  int i,j;
  int nb_write, offset, nb_color, nb_real_color, file_size, line_size;
  int first, last, x, color_ind, nb_byte;
  int palette_ind[257];
  unsigned char pal_bmp[256*4];
  unsigned char zero[4] = {0x00,0x00,0x00,0x00};
  unsigned char red, green, blue;
  unsigned char *clipboard;
  BITMAPFILEHEADER bmfh;
  BITMAPINFOHEADER bmih;
  
  /* Init */
  clipboard = (unsigned char *) calloc(1,height*width*3);
  if(clipboard == NULL)
     return(ERR_GRAPHIC_FOPEN);
  nb_byte = 3;
  nb_real_color = Compute8BitColors(nb_byte,height,width,picture,&(palette_ind[0]));
  
  /* Cration du fichier */
  fd = fopen(file_path,"wb");
  if(fd == NULL)
    return(ERR_GRAPHIC_FOPEN);

  /* Init */
  memset(&bmfh,0,sizeof(BITMAPFILEHEADER));
  memset(&bmih,0,sizeof(BITMAPINFOHEADER));
  memset(&pal_bmp,0,256*4);

  /** Cration de l'entte **/
  bmfh.bfType = 19778;  /* BM */
  bmih.biWidth = width;
  bmih.biHeight = height;
  bmih.biPlanes = 1;

  /*** Cration des points ***/
  if(nb_real_color <= 16)
    {
      /** Codage Palette 16 couleurs **/
      nb_color = 16;

      /* Entte */
      line_size = (width+1)/2;  /* Alignement 32 bit */
      if(line_size%4)
        line_size += (4-(line_size%4));
      bmfh.bfSize = 0;
      bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + nb_color*4;
      nb_write = fwrite(&bmfh,1,sizeof(BITMAPFILEHEADER),fd);
      if(nb_write != sizeof(BITMAPFILEHEADER))
        {
          fclose(fd);
          _unlink(file_path);
          return(ERR_GRAPHIC_WRITEFILE);
        }
      bmih.biSize = sizeof(BITMAPINFOHEADER);
      bmih.biBitCount = 4;
      bmih.biSizeImage = line_size*height;
      nb_write = fwrite(&bmih,1,sizeof(BITMAPINFOHEADER),fd);
      if(nb_write != sizeof(BITMAPINFOHEADER))
        {
          fclose(fd);
          _unlink(file_path);
          return(ERR_GRAPHIC_WRITEFILE);
        }

      /* Palette */
      for(i=0; i<nb_real_color; i++)
        {
          pal_bmp[i*4+2] = (palette_ind[i]&0x00FF0000)>>16;
          pal_bmp[i*4+1] = (palette_ind[i]&0x0000FF00)>>8;
          pal_bmp[i*4+0] = (palette_ind[i]&0x000000FF);
        }
      nb_write = fwrite(&pal_bmp,1,nb_color*4,fd);
      if(nb_write != nb_color*4)
        {
          fclose(fd);
          _unlink(file_path);
          return(ERR_GRAPHIC_WRITEFILE);
        }

      /** Points **/
      for(i=height-1; i>=0; i--)
        {
          offset = i*width*3;

          /* Prpare les points (rcupre l'indice) */
          for(j=0; j<width; j++)
            {
              red = picture[(j+i*width)*3];
              green = picture[(j+i*width)*3+1];
              blue = picture[(j+i*width)*3+2];
              color_ind = 65536*red + 256*green + blue;

              /** Recherche dichotomique de la valeur **/
              for(first=0,last=nb_real_color; last>=first;)
                {
                  x = (first+last)/2;
                  if(color_ind < palette_ind[x])
                    last = x - 1;
                  else
                    first = x + 1;

                  if(color_ind == palette_ind[x])
                    break;
                }
              clipboard[j/2]=(j%2)?clipboard[j/2]|(unsigned char)x:((unsigned char)x)<<4;
            }

          /* Points */
          nb_write = fwrite(clipboard,1,(width+1)/2,fd);
          if(nb_write != (int)((width+1)/2))
            {
              fclose(fd);
              _unlink(file_path);
              return(ERR_GRAPHIC_WRITEFILE);
            }

          /* Alignement (si besoin) */
          if(((width+1)/2)%4)
            {
              nb_write = fwrite(zero,1,4-(((width+1)/2)%4),fd);
              if(nb_write != (int)(4-(((width+1)/2)%4)))
                {
                  fclose(fd);
                  _unlink(file_path);
                  return(ERR_GRAPHIC_WRITEFILE);
                }
            }
        }
    }
  else if(nb_real_color <= 256)
    {
      /** Codage Palette 256 couleurs **/
      nb_color = 256;

      /* Entte */
      line_size = width;  /* Alignement 32 bit */
      if(line_size%4)
        line_size += (4-(line_size%4));
      bmfh.bfSize = 0;
      bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + nb_color*4;
      nb_write = fwrite(&bmfh,1,sizeof(BITMAPFILEHEADER),fd);
      if(nb_write != sizeof(BITMAPFILEHEADER))
        {
          fclose(fd);
          _unlink(file_path);
          return(ERR_GRAPHIC_WRITEFILE);
        }
      bmih.biSize = sizeof(BITMAPINFOHEADER);
      bmih.biBitCount = 8;
      bmih.biSizeImage = line_size*height;
      nb_write = fwrite(&bmih,1,sizeof(BITMAPINFOHEADER),fd);
      if(nb_write != sizeof(BITMAPINFOHEADER))
        {
          fclose(fd);
          _unlink(file_path);
          return(ERR_GRAPHIC_WRITEFILE);
        }

      /* Palette */
      for(i=0; i<nb_real_color; i++)
        {
          pal_bmp[i*4+2] = (palette_ind[i]&0x00FF0000)>>16;
          pal_bmp[i*4+1] = (palette_ind[i]&0x0000FF00)>>8;
          pal_bmp[i*4+0] = (palette_ind[i]&0x000000FF);
        }
      nb_write = fwrite(&pal_bmp,1,nb_color*4,fd);
      if(nb_write != nb_color*4)
        {
          fclose(fd);
          _unlink(file_path);
          return(ERR_GRAPHIC_WRITEFILE);
        }

      /** Points **/
      for(i=height-1; i>=0; i--)
        {
          offset = i*width*3;

          /** Prpare les points (rcupre l'indice) **/
          for(j=0; j<width; j++)
            {
              red = picture[(j+i*width)*3];
              green = picture[(j+i*width)*3+1];
              blue = picture[(j+i*width)*3+2];
              color_ind = 65536*red + 256*green + blue;

              /** Recherche dichotomique de la valeur **/
              for(first=0,last=nb_real_color; last>=first;)
                {
                  x = (first+last)/2;
                  if(color_ind < palette_ind[x])
                    last = x - 1;
                  else
                    first = x + 1;

                  if(color_ind == palette_ind[x])
                    break;
                }

              clipboard[j] = (unsigned char)x;
            }

          /* Points */
          nb_write = fwrite(clipboard,1,width,fd);
          if(nb_write != (int)(width))
            {
              fclose(fd);
              _unlink(file_path);
              return(ERR_GRAPHIC_WRITEFILE);
            }

          /* Alignement (si besoin) */
          if((width)%4)
            {
              nb_write = fwrite(zero,1,4-(width%4),fd);
              if(nb_write != (int)(4-(width%4)))
                {
                  fclose(fd);
                  _unlink(file_path);
                  return(ERR_GRAPHIC_WRITEFILE);
                }
            }
        }
    }
  else
    {
      /** Codage True Color 24 bit **/
      /* Entte */
      /* Repre relatif pour +256 couleurs */
      nb_color = 257;
      line_size = width*3;  /* Alignement 32 bit */
      if(line_size % 4)
        line_size += (4-(line_size%4));
      bmfh.bfSize = 0;
      bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
      nb_write = fwrite(&bmfh,1,sizeof(BITMAPFILEHEADER),fd);
      if(nb_write != sizeof(BITMAPFILEHEADER))
        {
          fclose(fd);
          _unlink(file_path);
          return(ERR_GRAPHIC_WRITEFILE);
        }
      bmih.biSize = sizeof(BITMAPINFOHEADER);
      bmih.biBitCount = 24;
      bmih.biSizeImage = line_size*height;
      nb_write = fwrite(&bmih,1,sizeof(BITMAPINFOHEADER),fd);
      if(nb_write != sizeof(BITMAPINFOHEADER))
        {
          fclose(fd);
          _unlink(file_path);
          return(ERR_GRAPHIC_WRITEFILE);
        }

      /** Points **/
      for(i=height-1; i>=0; i--)
        {
          offset = i*width*3;

          /* Prpare les points (inverse les composantes RGB) */
          for(j=0; j<width; j++)
            {
              clipboard[j*3+0] = picture[offset+j*3+2];
              clipboard[j*3+1] = picture[offset+j*3+1];
              clipboard[j*3+2] = picture[offset+j*3+0];
            }

          /* Points */
          nb_write = fwrite(clipboard,1,width*3,fd);
          if(nb_write != (int)(width*3))
            {
              fclose(fd);
              _unlink(file_path);
              return(ERR_GRAPHIC_WRITEFILE);
            }

          /* Alignement (si besoin) */
          if((width*3)%4)
            {
              nb_write = fwrite(zero,1,4-((width*3)%4),fd);
              if(nb_write != (int)(4-((width*3)%4))) 
                {
                  fclose(fd);
                  _unlink(file_path);
                  return(ERR_GRAPHIC_WRITEFILE);
                }
            }
        }
    }

  /** Il faut indiquer la taille du fichier en entte **/
  fseek(fd,0L,SEEK_END);
  file_size = ftell(fd);
  fseek(fd, 0L, SEEK_SET);
  bmfh.bfSize = file_size;
  nb_write = fwrite(&bmfh,1,sizeof(BITMAPFILEHEADER),fd);
  if(nb_write != sizeof(BITMAPFILEHEADER))
    {
      fclose(fd);
      _unlink(file_path);
      return(ERR_GRAPHIC_WRITEFILE);
    }
  fseek(fd,0L,SEEK_END);

  /* Fermeture du fichier */
  fclose(fd);
    
  /* Libration mmoire */
  free(clipboard);

  /* Ok */
  return(0);
}


/**********************************************************/
/*  Compute8BitColors() : Compte les couleurs de l'image. */
/**********************************************************/
int Compute8BitColors(int nb_byte, int height, int width, unsigned char *windows_bitmap_data, int *palette_ind)
{
  unsigned char red, green, blue;
  int i, j, first, last, x, color_ind, window_height, window_width, nb_real_color, offset;
  
  /* Init */
  memset(palette_ind,0,257*sizeof(int));
  window_height = height;
  window_width = width;
  nb_real_color = 0;

  /** Passe tous les points en revue **/
  for(j=0; j<window_height && nb_real_color != 257; j++)
    for(i=0; i<window_width && nb_real_color != 257; i++)
      {
        /* Couleur du point */
        offset = (i+j*window_width)*nb_byte;
        red = windows_bitmap_data[offset];
        green = windows_bitmap_data[offset+1];
        blue = windows_bitmap_data[offset+2];
        color_ind = 65536*red + 256*green + blue;

        /** On insert/recherche la couleur (dichotomie) **/
        if(i == 0 && j == 0)
          {
            palette_ind[0] = color_ind;
            nb_real_color = 1;
          }
        else
          {
            for(first=0,last=nb_real_color; last>=first;)
              {
                x = (first+last)/2;
                if(color_ind < palette_ind[x])
                  last = x - 1;
                else
                  first = x + 1;

                if(color_ind == palette_ind[x])
                  break;
               }

            /* A t'on trouv la couleur ou doit on l'ajouter */
            if(color_ind != palette_ind[x])
              {
                /* Dcalage du tableau */
                if(x < nb_real_color)
                  if(palette_ind[x] < color_ind)
                    x++;
                memmove(&palette_ind[x+1],&palette_ind[x],(nb_real_color-x)*sizeof(int));
                palette_ind[x] = color_ind;
                nb_real_color++;
              }
          }
      }

  /* Renvoi le nombre de couleurs */
  return(nb_real_color);
}

/***************************************************************************************/